package liv;

import java.io.FileInputStream;
import java.util.concurrent.TimeUnit;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Test the speed efficiency of xpath.
 *
 * @author Liviu Tudor http://about.me/liviutudor
 */
public class XPathTest {
    public static final String ATTR_NAME        = "name";
    public static final String ATTR_DESCRIPTION = "description";
    /**
     * Stores the xml filename to test against. This is passed in the command
     * line.
     */
    private String             filename;

    /**
     * Stores the DOM (parsed in {@link #parseFile()}.
     */
    private NodeList           nodes;

    private DomProcessing[]       proc;

    /**
     * XPath engine.
     */
    private XPath              xpath;

    /**
     * XPath expression for selecting the nodes we need.
     */
    private XPathExpression    xpathItem;

    /**
     * XPath expression for extracting the attribute {@link #ATTR_NAME}.
     */
    private XPathExpression    xpathAttrName;
    /**
     * XPath expression for extracting the attribute {@link #ATTR_DESCRIPTION}.
     */
    private XPathExpression    xpathAttrDesc;

    /**
     * Constructor which creates a test for the given XML file.
     *
     * @param filename
     *            Name of the XML file to test against.
     */
    public XPathTest(String filename) throws Exception {
        this.filename = filename;
        XPathFactory xfactory = XPathFactory.newInstance();
        xpath = xfactory.newXPath();
        xpathItem = xpath.compile("/data/item");
        xpathAttrName = xpath.compile("@" + ATTR_NAME);
        xpathAttrDesc = xpath.compile("@" + ATTR_DESCRIPTION);
        proc = new DomProcessing[] { new DomProcessingXPath(), new DomProcessingAttr() };
    }

    /**
     * Program entry point -- supports only one 1 parameter: the filename.
     *
     * @param args
     *            Command line parameters
     */
    public static void main(String[] args) {
        if (args.length != 1) {
            System.out.println("Usage: XPathTest <filename>");
            System.exit(1);
        }
        try {
            XPathTest test = new XPathTest(args[0]);
            test.performTest();
        } catch (Exception e) {
            System.out.println("Error occur during execution:" + e.getMessage());
            System.exit(2);
        }
    }

    /**
     * Parses the document and extracts the <code>item</code> nodes.
     *
     * @throws Exception
     *             if any I/O errors occur or no nodes are found.
     */
    public void parseFile() throws Exception {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(false);
        DocumentBuilder builder = factory.newDocumentBuilder();
        Document doc = builder.parse(new FileInputStream(filename));
        // having parsed the document, run it through xpath and get the nodes
        Object res = xpathItem.evaluate(doc, XPathConstants.NODESET);
        if (res == null) throw new Exception("No nodes found!");
        nodes = (NodeList) res;
    }

    public void performTest() throws Exception {
        // parse the file first
        parseFile();
        // ensure we remove anything from memory we don't need
        System.gc();
        TimeUnit.SECONDS.sleep(5); // give the system time to do whatever is
                                   // needed to cleanup
        //finally iterate through this and measure
        int steps = 10;  //perform each step 10 times to average
        for( int i = 0; i < proc.length; i++ ) {
            System.out.println( "\nProcessor " + i );
            for( int j = 0; j < steps; j++ ) {
                long start = System.nanoTime();
                traverse( proc[i] );
                long end = System.nanoTime();
                System.out.println( end - start );
            }
        }
    }

    public void traverse( DomProcessing p ) throws Exception {
        @SuppressWarnings("unused")
        Item itm = null;
        for( int i = 0; i < nodes.getLength(); i++ ) {
            Node n = nodes.item(i);
            itm = p.process((Element)n);
        }
    }

    /**
     * Encapsulates an "item" from our XML feed. Such an "item" only has a name
     * and a description.
     *
     * @author Liviu Tudor http://about.me/liviutudor
     */
    class Item {
        private String name;
        private String description;

        public Item() {
            this(null, null);
        }

        public Item(String name, String description) {
            this.name = name;
            this.description = description;
        }

        String getName() {
            return name;
        }

        void setName(String name) {
            this.name = name;
        }

        String getDescription() {
            return description;
        }

        void setDescription(String description) {
            this.description = description;
        }

        @Override
        public String toString() {
            return "Name=" + name + ", Description=" + description;
        }
    }

    /**
     * Interface which defines a DOM processing. Simply given a DOM Element it
     * marshals that into an {@link Item} which encapsulates the same data.
     *
     * @author Liviu Tudor http://about.me/liviutudor
     */
    interface DomProcessing {
        Item process(Element e) throws Exception;
    }

    /**
     * Uses XPath expressions to retrieve attributes and create an Item.
     *
     * @author Liviu Tudor http://about.me/liviutudor
     */
    class DomProcessingXPath implements DomProcessing {
        @Override
        public Item process(Element e) throws Exception {
            String name = (String) xpathAttrName.evaluate(e, XPathConstants.STRING);
            String desc = (String) xpathAttrDesc.evaluate(e, XPathConstants.STRING);
            return new Item(name, desc);
        }
    }

    /**
     * Uses regular DOM <code>getAttribute</code> to retrieve attributes and
     * create an Item.
     *
     * @author Liviu Tudor http://about.me/liviutudor
     */
    class DomProcessingAttr implements DomProcessing {
        @Override
        public Item process(Element e) throws Exception {
            String name = e.getAttribute(ATTR_NAME);
            String desc = e.getAttribute(ATTR_DESCRIPTION);
            return new Item(name, desc);
        }
    }
}
